import java.util.ArrayList;
import java.util.concurrent.locks.*;

public class SimpleConcurrentQueue<E> implements SimpleBlockingQueueInterface<E> {

  // The queue has bounded capacity.
  final int capacity;
  // This array holds the queue elements.
  private final ArrayList<E> elements;
  // The number of the next element to be dequeued.
  private int head;
  // The number of the next element to be inserted.
  private int tail;
  // The lock that protects this queue.
  private final Lock lock = new ReentrantLock ();
  // The condition variables that we use to wait for the queue to
  // become non-empty, or non-full, respectively.
  private final Condition notEmpty = lock.newCondition();
  private final Condition notFull  = lock.newCondition();

  // Initialization.

  public SimpleConcurrentQueue (int capacity)
  {
    assert (capacity > 0);
    this.capacity = capacity;
    this.elements = new ArrayList<E> (capacity);
    for (int i = 0; i < capacity; i++)
      this.elements.add(null);
    this.head = this.tail = 0;
  }

  // Inserting an element into the queue.

  public void put (E element)
  {
    // Acquire the lock.
    lock.lock();
    try {

      // We need the queue not to be full. If this condition is not
      // met now, we abandon the lock, suspend until someone signals
      // that this condition might have become true, re-acquire the
      // lock, and check again.
      while (head + capacity == tail)
	notFull.awaitUninterruptibly();

      // At this point, we hold the lock, and we have tail < head +
      // capacity, so the slot at index (tail % capacity) is
      // available.

      // Note whether the queue is empty.
      boolean wasEmpty = (head == tail);

      // Write the new element into this slot.
      elements.set(tail % capacity, element);
      // Mark this slot as occupied.
      tail++;
      // We again have tail <= head + capacity.

      // If the queue has just transitioned from empty to non-empty,
      // wake up ALL threads that were blocked in take().
      if (wasEmpty)
	notEmpty.signalAll();
      
    } finally {
      lock.unlock();
    }
  }

  // Extracting an element out of the queue.

  public E take ()
  {
    // Acquire the lock.
    lock.lock();
    try {

      // We need to queue to be non-empty. If this condition is not
      // met now, we abandon the lock, suspend until someone signals
      // that this condition might have become true, re-acquire the
      // lock, and check again.
      while (head == tail)
	notEmpty.awaitUninterruptibly();

      // At this point, we hold the lock, and we have head < tail,
      // so the slot at index (head % capacity) is occupied.

      // Note whether the queue is full.
      boolean wasFull = (head + capacity == tail);

      // Read the element in this slot.
      E element = elements.get(head % capacity);
      // Erase this slot. Although not strictly necessary in
      // principle, this instruction ensures that any non-occupied
      // slot contains the value null. This is required in order to
      // avoid a memory leak.
      elements.set(head % capacity, null);
      // Mark this slot as free.
      head++;
      // We again have head <= tail.
	    
      // If the queue has just transitioned from full to non-full,
      // wake up ALL threads that were blocked in put().
      if (wasFull)
	notFull.signalAll();
      
      // Return the element that has been read.
      return element;

    } finally {
      lock.unlock();
    }
  }

}

/* Personal note:

   Writing "new E [capacity]" produces an error in Java ("generic array creation").

   One can write "(E[]) new Object [capacity]". This is not nice, because it
   produces a warning, and indeed it is potentially unsafe. It is safe only if
   one can convince oneself that only values of type E will ever be written to
   this array (indeed, creating the array at type Object means that Java will
   perform no check when writing elements into it; so elements of type Object
   *can* be written to this array).  It would be safe here.

   One can also write "(E[]) java.lang.reflect.Array.newInstance(c, capacity)".
   This is not convenient, because one must pass the parameter c, of type
   Class<E>, at runtime; plus it still requires a cast! I don't understand
   why newInstance does not return an array of type E[].

   Perhaps the most elegant solution is to just use Vector -- though these
   are resizable arrays, so they are overkill, who cares? Ah, the trouble is,
   Vector is synchronized, so it does not make much sense to use Vector in
   this particular application... Argh! It seems that ArrayList is roughly
   equivalent to Vector and unsynchronized. Is this the solution?

*/

