import java.util.ArrayList;
import java.util.concurrent.locks.*;

public class SimplisticQueue<E>
  implements SimpleNonBlockingQueueInterface<E> {

  // The queue has bounded capacity.
  final int capacity;
  // This array holds the queue elements.
  private final ArrayList<E> elements;
  // The number of the next element to be dequeued.
  private int head;
  // The number of the next element to be inserted.
  private int tail;

  // Initialization.

  public SimplisticQueue (int capacity)
  {
    assert (capacity > 0);
    this.capacity = capacity;
    this.elements = new ArrayList<E> (capacity);
    for (int i = 0; i < capacity; i++)
      this.elements.add(null);
    this.head = this.tail = 0;
  }

  // Inserting an element into the queue.

  public void put (E element) throws ImpossibleNow
  {
    // We need the queue not to be full. If this condition
    // is not met now, we abort the operation.
    if (head + capacity == tail)
      throw new ImpossibleNow ();
      
    // We have tail < head + capacity, so the slot at index
    // (tail % capacity) is available.

    // Write the new element into this slot.
    elements.set(tail % capacity, element);
    // Mark this slot as occupied.
    tail++;
    // We again have tail <= head + capacity.
  }

  // Extracting an element out of the queue.

  public E take () throws ImpossibleNow
  {
    // We need to queue to be non-empty. If this condition
    // is not met now, we abort the operation.
    if (head == tail)
      throw new ImpossibleNow ();

    // We have head < tail, so the slot at index
    // (head % capacity) is occupied.

    // Read the element in this slot.
    E element = elements.get(head % capacity);
    // Erase this slot. Although not strictly necessary in
    // principle, this instruction ensures that any
    // non-occupied slot contains the value null. This is
    // required in order to avoid a memory leak.
    elements.set(head % capacity, null);
    // Mark this slot as free.
    head++;
    // We again have head <= tail.
      
    // Return the element that has been read.
    return element;

  }

}
